#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <sndfile.h>
#include <math.h>

unsigned char audio[1024*1024];
unsigned char code[32768];
uint16_t offset[49];

#define BIT(x,n) (((x)>>(n))&1)

#define BITSWAP16(val,B15,B14,B13,B12,B11,B10,B9,B8,B7,B6,B5,B4,B3,B2,B1,B0) \
        ((BIT(val,B15) << 15) | (BIT(val,B14) << 14) | (BIT(val,B13) << 13) | (BIT(val,B12) << 12) | \
                (BIT(val,B11) << 11) | (BIT(val,B10) << 10) | (BIT(val, B9) <<  9) | (BIT(val, B8) <<  8) | \
                (BIT(val, B7) <<  7) | (BIT(val, B6) <<  6) | (BIT(val, B5) <<  5) | (BIT(val, B4) <<  4) | \
                (BIT(val, B3) <<  3) | (BIT(val, B2) <<  2) | (BIT(val, B1) <<  1) | (BIT(val, B0) <<  0))

void load_code(char *file)
{
  unsigned char cc[32768];
  int i;
  FILE *f = fopen(file, "rb"); if (!f) { perror(file); exit(1); }
  fread(cc, 32768, 1, f);
  fclose(f);
  for (i = 0; i < 32768; i++) {
    code[BITSWAP16(i,15,14,13,12,11,10,9,8,0,1,2,3,4,5,6,7)] = cc[i];
  }
}

void save_code(char *file)
{
  unsigned char cc[32768];
  int i;
  FILE *f;
  for (i = 0; i < 32768; i++) {
    cc[i] = code[BITSWAP16(i,15,14,13,12,11,10,9,8,0,1,2,3,4,5,6,7)];
  }
  f = fopen(file, "wb"); if (!f) { perror(file); exit(1); }
  fwrite(cc, 32768, 1, f);
  fclose(f);
}

void usage(void)
{
  printf("gimme <rom code> <rom audio 1> <rom audio 2> <sample 1> ... <sample N>\n");
  printf("N is 49 at max\n");
  exit(1);
}

void put_audio(long pos, unsigned char v)
{
  if (pos >= 1024*1024) { printf("error: ROM full\n"); exit(1); }
  audio[pos] = v;
}

long load_sample(char *name, long loadpos)
{
  SF_INFO inf;
  SNDFILE *f;
  float *buf;
  float *max;
  sf_count_t c;
  int i;
  float gain;

  inf.format = 0;
  f = sf_open(name, SFM_READ, &inf);
  if (f == NULL) { printf("error opening %s\n", name); exit(1); }
  if (inf.channels != 1) {
    printf("%s: %d channels, only 1 is supported\n", name, inf.channels);
    exit(1);
  }
  printf("INFO: %s is at %d Hz (stored at 0x%4.4lx)\n", name, inf.samplerate,
         loadpos);

  buf = malloc(sizeof(float) * inf.frames); if (buf == NULL) abort();
  max = malloc(sizeof(float) * inf.frames); if (max == NULL) abort();

  c = sf_read_float(f, buf, inf.frames);
  if (c != inf.frames) { printf("%s: read error\n", name); exit(1); }
  sf_close(f);

  max[inf.frames-1] = fabsf(buf[inf.frames-1]);
  for (i = inf.frames-2; i >= 0; i--) {
    if (max[i+1] > fabsf(buf[i])) max[i] = max[i+1];
    else                          max[i] = fabsf(buf[i]);
  }

  gain = 1;
  for (i = 0; i < inf.frames-1; i++) {
    float v;
    max[i] *= gain;
    if (max[i] <= 0.5 && gain != 256)
      { put_audio(loadpos, 0x80); loadpos++; gain *= 2; }
    v = buf[i] * gain * 127;
    if (v < -127) v = -127;
    if (v > 127)  v = 127;
    put_audio(loadpos, lrint(v));
    loadpos++;
  }

  /* end marker */
  put_audio(loadpos, 0x80); loadpos++;
  put_audio(loadpos, 0x80); loadpos++;
  put_audio(loadpos, 0x80); loadpos++;
  while (loadpos & 15) { put_audio(loadpos, 0x80); loadpos++; }

  return loadpos;
}

int main(int n, char **v)
{
  char *samples[49];
  char *romcode = NULL;
  char *romaudio1 = NULL, *romaudio2 = NULL;
  int i;
  long loadpos;
  FILE *f;

  for (i = 0; i < 49; i++) samples[i] = NULL;

  if (n < 1 + 2 || n > 1 + 2 + 49) usage();
  romcode = v[1];
  romaudio1 = v[2];
  romaudio2 = v[3];
  for (i = 4; i < n; i++) {
    if (i - 4 >= 49) usage();
    samples[i-4] = v[i];
  }

  /* initial audio data: a 0 and an end marker */
  audio[1] = audio[2] = audio[3] = 0x80;

  /* load samples */
  loadpos = 0;
  for (i = 0; i < 49; i++) {
    if (samples[i] == NULL) break;
    offset[i] = loadpos >> 4;
    loadpos += load_sample(samples[i], loadpos);
  }

  /* save audio files */
  f = fopen(romaudio1, "wb"); if (f == NULL) { perror(romaudio1); exit(1); }
  fwrite(audio, 512*1024, 1, f);
  fclose(f);
  f = fopen(romaudio2, "wb"); if (f == NULL) { perror(romaudio2); exit(1); }
  fwrite(audio+512*1024, 512*1024, 1, f);
  fclose(f);

  /* update ROM */
  load_code(romcode);
  for (i = 0; i < 49; i++) {
    unsigned char *b = code + 0x6bfd + i * 2;
    b[0] = (offset[i] >> 8) & 255;
    b[1] = offset[i] & 255;
  }
  save_code(romcode);

  return 0;
}
